**Table Of Contents**
___

[[_toc_]]
___

This tutorial is largely inspired by the [CuPy documentation](https://docs.cupy.dev/en/stable/index.html)

___

# 1. Overview of library CuPy


## 1.1 Project goal

[Ref Cupy](https://docs.cupy.dev/en/stable/overview.html#project-goal)

> The goal of the CuPy project is to provide Python users GPU acceleration capabilities, without the in-depth knowledge of underlying GPU technologies. The CuPy team focuses on providing:

> 1) A complete NumPy and SciPy API coverage to become a full drop-in replacement, as well as advanced CUDA features to maximize the performance.

> 2) Mature and quality library as a fundamental package for all projects needing acceleration, from a lab environment to a large-scale cluster.


## 1.2 Backend for GPU NVidia

[Ref Cupy](https://cupy.dev/)

> CuPy utilizes CUDA Toolkit libraries including:
> * cuBLAS,
> * cuRAND,
> * cuSOLVER,
> * cuSPARSE,
> * cuFFT,
> * cuDNN,
> * NCCL

> to make full use of the GPU architecture.

## 1.3 Backend for GPU AMD

[Ref Cupy](https://docs.cupy.dev/en/stable/install.html#using-cupy-on-amd-gpu-experimental)

* For AMD GPU CuPy has an experimental support with ROCm **v4.3 / v5.0**
* With also some limitations like feature specific to CUDA

## 1.4 Speedup over Numpy

![alt text](img/speedup_cupy.png)

[Ref Cupy](https://cupy.dev/)

Speed up depends of :

* the size of data, **small data results in less gain**
* type of processing, 
  * **element wise operations bring the best gain**
  * There can be a factor of 10 between the speedup for a complex operation and an element wise operation, speedup is in log scale

Remark:
  * noted the difference in speedup between element-by-element operations $O(n)$ and a matrix multiplication vector $O(n^2)$
  * matrix multiplication vector must combined elements "distant" in memory, the acceleration does not increase with the size of the matrix, ie data as for other operations

# 2. Array and memory

DEVICE and HOST have their own memory space,

* The term DEVICE is used to designate the **G**PU
* The term HOST is used to designate the **C**PU

 but we are moving towards an architecture where CPU and GPU use a share memory, new NVidia "Grace Hopper SuperChip".

## 2.1 HOST memory and `numpy.ndarray`

* Numpy arrays are obviously stored in the HOST memory linked with the CPU

## 2.2 DEVICE memory and `cupy.ndarray`

* `cupy.ndarray` return an array stored in DEVICE memory 
* All array allocation is done on 'Current device' by default with ID 0

```python
import cupy as cp

x_gpu = cp.array([1, 2, 3])
x_gpu.device
```
Result
```python
<CUDA Device 0>
```

## 2.3 Data transfer

[Ref CuPy](https://docs.cupy.dev/en/stable/user_guide/basic.html#data-transfer)

### HOST to DEVICE

* With cupy function `cupy.asarray(host_array)` 

```python
import cupy as cp
import numpy as np

x_cpu = np.array([1, 2, 3])

x_gpu = cp.asarray(x_cpu)  # move the data to the current device.
```

### DEVICE to HOST

* With method `get()` of the CuPy array object
```python
import cupy as cp
import numpy as np

x_gpu = cp.array([1, 2, 3])

x_CPU = x_gpu.get()
```
* With numpy function `numpy.asnumpy(my_a)`

```python
import cupy as cp
import numpy as np

x_gpu = cp.array([1, 2, 3])

x_CPU = np.asnumpy(x_gpu)
```

### Warning : cost of transfer

>Note that converting between cupy.ndarray and numpy.ndarray incurs data transfer between the host (CPU) device and the GPU device, which is costly in terms of performance.

[Ref CuPy](https://docs.cupy.dev/en/stable/reference/ndarray.html#conversion-to-from-numpy-arrays)

* Small, frequent memory transfers penalize performance
* Good memory usage and transfers are the key to good CPU performance.


# 3. Numpy and Scipy clones 

## 3.1 CuPy array

* All routines of Numpy and Scipy 'GPU clone' applied to an object `cupy.ndarray`
* An array `cupy.ndarray` is very similar to `numpy.ndarray`: 
  * same attribut and more
    * `device` contents the ID of DEVICE where array is stored
  * same method with a few exceptions
    * `resize()` not available
    * `get()` return a  `numpy.ndarray` object
  * same slicing features, but
    * Cupy never returns **IndexError** integer array indexing, see below
  * broadcasting 
  * dtype,  exception: float128 not available
* ... How to distinguish them then, with `type`

```python
isinstance(x_gpu, cp.ndarray)
True
```
### Out-of-bounds indices

```python
In [24]: x_gpu
Out[24]: array([1, 2, 3])

In [25]: x_gpu[10]
---------------------------------------------------------------------------
IndexError                                Traceback (most recent call last)
Cell In[25], line 1
----> 1 x_gpu[10]

File cupy/_core/core.pyx:1521, in cupy._core.core._ndarray_base.__getitem__()

IndexError: Index 10 is out of bounds for axis 0 with size 3
```

but 
```python
In [26]: x_gpu[[1,10]]
Out[26]: array([2, 2])
```
produces no error.

## 3.2 Routines cupy.numpy

### Coverage of routines

See [comparaison table](https://docs.cupy.dev/en/stable/reference/comparison.html#numpy-cupy-apis)

### Some differences

See Cupy documentation [Differences between CuPy and NumPy](https://docs.cupy.dev/en/stable/user_guide/difference.html#)

Those that seem important to me

* Random methods support dtype argument
  * Cupy support dtype argument in random function, Numpy not always float64

* Zero-dimensional array
  * Cupy returns a  zero-dimensional `cupy.ndarray` for scalar value
* Structured arrays and record arrays are currently unsupported.

### First example of processing on GPU

```python
import cupy as cp

def my_func(a):
    return cp.sqrt(cp.sum(a**2, axis=-1))

a = cp.random.random((2048, 1024))
res = my_func(a)
res.shape
(2048,)
```


### Benchmark tools


[Ref CuPy](https://docs.cupy.dev/en/stable/user_guide/performance.html#benchmarking)

>CuPy provides a useful utility `cupyx.profiler.benchmark()` for timing the elapsed time of a Python function on both CPU and GPU:

```python
from cupyx.profiler import benchmark

print(benchmark(my_func, (a,), n_repeat=20)) 
```
Result
```python
my_func             :    
CPU:   195.126 us   +/- 10.652 (min:   185.952 / max:   233.408) us     
GPU-0:   205.261 us   +/- 11.064 (min:   194.560 / max:   243.712) us
```
Global time is 400 $\mu s$

Compare with 100% CPU execution

```python
import numpy as np

def func_numpy(a):
    return np.sqrt(np.sum(a**2, axis=-1)) 

b=a.get()

%timeit func_numpy(b)
3.98 ms ± 10.7 μs per loop (mean ± std. dev. of 7 runs, 100 loops each)
```
4 ms for CPU, GPU speedup **x10**

Compare now the result

```python
res_np = func_numpy(b)

np.allclose(res.get(), res_np)
True
```

## 3.3 Routines cupy.scipy

### Somes differences

See [comparaison table](https://docs.cupy.dev/en/stable/reference/comparison.html#scipy-cupy-apis)

* Best coverage :  [multidimensional-image-processing](https://docs.cupy.dev/en/stable/reference/comparison.html#multidimensional-image-processing)
* Worst coverage :  [statistical-functions](https://docs.cupy.dev/en/stable/reference/comparison.html#statistical-functions)


## 3.4 The important points

* CuPy arrays are very close to numpy array
* An conversion between CuPy Numpy array causes a physical transfer between CPU/GPU memory
* a CuPy function of Numpy or Scipy are processing on GPU 


# 4. User defined kernels

Here the list of the kind kernel users 

* https://docs.cupy.dev/en/stable/reference/kernel.html

* https://docs.cupy.dev/en/stable/user_guide/kernel.html#user-defined-kernels

we present 3

## 4.1 RawKernel() : like C CUDA


>RawKernel object allows you to call the kernel with CUDA’s cuLaunchKernel interface. In other words, you have control over grid size, block size, shared memory size and stream.*

[Ref Doc](https://docs.cupy.dev/en/stable/user_guide/kernel.html#raw-kernels)

### Code example : 

```python
add_kernel = cp.RawKernel(r'''
extern "C" __global__
void my_add(const float* x1, const float* x2, float* y) {
    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    y[tid] = x1[tid] + x2[tid];
}
''', 'my_add')
x1 = cp.arange(25, dtype=cp.float32).reshape(5, 5)
x2 = cp.arange(25, dtype=cp.float32).reshape(5, 5)
y = cp.zeros((5, 5), dtype=cp.float32)
add_kernel((5,), (5,), (x1, x2, y))  # grid, block and arguments
y
array([[ 0.,  2.,  4.,  6.,  8.],
       [10., 12., 14., 16., 18.],
       [20., 22., 24., 26., 28.],
       [30., 32., 34., 36., 38.],
       [40., 42., 44., 46., 48.]], dtype=float32)
```

## 4.2 jit.rawkernel()

Close to `RawKernel()` but with python syntax, so you have already control over grid size, block size ...

But
>CuPy’s array operations which automatically parallelize operations (e.g., add(), sum()) are not supported.

### Code example : 
```python
from cupyx import jit

@jit.rawkernel()
def elementwise_copy(x, y, size):
    tid = jit.blockIdx.x * jit.blockDim.x + jit.threadIdx.x
    ntid = jit.gridDim.x * jit.blockDim.x
    for i in range(tid, size, ntid):
        y[i] = x[i]
```

```python
size = cupy.uint32(2 ** 22)
x = cupy.random.normal(size=(size,), dtype=cupy.float32)
y = cupy.empty((size,), dtype=cupy.float32)

elementwise_copy((128,), (1024,), (x, y, size))  # RawKernel style
assert (x == y).all()

elementwise_copy[128, 1024](x, y, size)  #  Numba style
assert (x == y).all()
```

## 4.3 Pythonic kernel with `fuse()`

>`cupy.fuse()` is a decorator that fuses (pure python) functions. This decorator can be used to define an elementwise or reduction kernel. 

### Code example : 

```python
@cupy.fuse(kernel_name='squared_diff')
def squared_diff(x, y):
    return (x - y) * (x - y)
```

```python
x = cupy.arange(10)
y = cupy.arange(10)[::-1]
squared_diff(x, y)
array([81, 49, 25,  9,  1,  1,  9, 25, 49, 81])
```

You can use condition instruction and call numpy function

```python
@cp.fuse()
def sum_of_products(x, y):
    return cp.sum(x * y, axis = -1)
```

Limitation:

* >Most other routines (e.g. `cupy.matmul()`, `cupy.reshape()`) are not supported.
* You **can't control** grid size, block size, shared memory size and stream like for `RawKernel`


## 4.4 Discussion on the scope of the course Python GPU
 
The use of raw kernels is covered by other languages ​​(C++/RUST/FORTRAN) in Gray-Scott School, so in this course we will only examine kernel fusion `cupy.fuse()` and the exploitation of CuPy versions of Numpy and Scipy functions.

However, keep in mind that it is possible to have control over grid size, block size, shared memory size and stream with CuPy and also stream synchronization , event ... see [Concurrent access to the GPU](https://carpentries-incubator.github.io/lesson-gpu-programming/instructor/streams.html) with CuPy.

Others low level libraries for GPU:
* [Numba for CUDA GPUs](https://numba.pydata.org/numba-doc/latest/cuda/index.html)
* [PyCUDA](https://documen.tician.de/pycuda/)


# 5. Gray-Scott equation

Now we have all the CuPy knowledge to solve the Gray-Scott reaction-diffusion equation with the finite difference method.

## 5.1 Finite difference method

See [Numpy Course](https://gitlab.in2p3.fr/alice.faure/gray-scott-python/-/blob/main/tutorial/1_Python_CPU_Numpy.ipynb?ref_type=heads)

### Step 1/3 : processing the laplacian

We can compute laplacian of u and v with a convolution with stencil 

**Exercice:**

Write 2D convolution between array image and laplacian stencil 3x3 with function `convolve2d()` in `cupyx.scipy.signal` module


**Solution**

<details>
<summary>Click to expand</summary>

```python
import cupy as cp
from cupyx.scipy.signal import convolve2d as conv2d_gpu

stencil = cp.array([[0, 1, 0],
                    [1, -4, 1],
                    [0, 1, 0]],
                   dtype=cp.float32)
# Define U with numpy array
u_gpu = cp.asarray(U)
Lu = conv2d_gpu(u_gpu, stl_gpu, 'same', 'fill', 0)
```

</details>


### Step 2/3 : update time with kernel fusion

**Exercice:**

Write the Gray-Scott CuPy kernel function with this signature 

 `grayscott_kernel(Lu, Lv, u, v, Du, Dv, F, k, delta_t)`



**Solution**

<details>
<summary>Click to expand</summary>

```python
import cupy as cp

@cp.fuse()
def grayscott_kernel(Lu, Lv, u, v, Du, Dv, F, k, delta_t):
    uvv = u * v * v
    u += delta_t * ((Du * Lu - uvv) + F * (1 - u))
    v += delta_t * ((Dv * Lv + uvv) - v * (F + k))

```
</details>


### Step 3/3 : insert step 1, 2 in loop

**Exercice:**

Now insert the previous steps in a nested loop, and store in a numpy array `frames_V` an image all `step_frame` iterations, like

```python
for idx_fr  in range(nb_frame):
    for _ in range(step_frame):
        Lu = laplacian(u)
        Lv = laplacian(v)
        grayscott_kernel(Lu, Lv, u_gpu, v_gpu, Du, Dv, F, k, delta_t)
    frames_V[idx_fr] = v_gpu.get()
```

create a function with this signature :

`grayscott_cupy(U, V, Du, Dv, F, k, delta_t, nb_frame, step_frame)`

**Solution**

<details>
<summary>Click to expand</summary>

```python
import cupy as cp
import numpy as np 
from cupyx.scipy.signal import convolve2d as conv2d_gpu


@cp.fuse()
def grayscott_kernel(Lu, Lv, u, v, Du, Dv, F, k, delta_t):
    uvv = u * v * v
    u += delta_t * ((Du * Lu - uvv) + F * (1 - u))
    v += delta_t * ((Dv * Lv + uvv) - v * (F + k))


def grayscott_cupy(U, V, Du, Dv, F, k, delta_t, nb_frame, step_frame):
    '''
    Solve Gray-Scott equation with finite difference,
    hardward target GPU wiht cupy library :
    
       * laplacian is computed with 2D convolution provided by cupy
       * update U, V with elementwise GPU kernel
    '''
    # output video frames
    n_x, n_y = U.shape[0], U.shape[1]
    frames_V = np.empty((nb_frame, n_x, n_y), dtype=np.float32)
    # Laplacian stencil
    stencil = np.array([[0, 1, 0],
                        [1, -4, 1],
                        [0, 1, 0]],
                        dtype=np.float32)
    # Load in GPU memory
    stl_gpu = cp.array(stencil)
    u_gpu = cp.asarray(U)
    v_gpu = cp.asarray(V)
    for idx_fr  in range(nb_frame):
        for _ in range(step_frame):
            # compute laplacians with convolution provided by cupy
            Lu = conv2d_gpu(u_gpu, stl_gpu, 'same', 'fill', 0)
            Lv = conv2d_gpu(v_gpu, stl_gpu, 'same', 'fill', 0)
            grayscott_kernel(Lu, Lv, u_gpu, v_gpu, Du, Dv, F, k, delta_t)
        frames_V[idx_fr] = v_gpu.get()
    return frames_V
```

Main can be with package `compute.common`

```python
import compute.common as gsc

if __name__ == "__main__":
    U, V, _ = gsc.grayscott_init(1920, 1080)
    gs_pars = gsc.grayscott_pars()
    nb_frame = 100
    frames_ui = gsc.grayscott_main(grayscott_cupy, gs_pars, U, V, nb_frame)

```
</details>

## 5.2 Profilage and optimization

### interactive session avec GPU: SLURM and CONDOR

**CCIN2P3** (at Lyon) with schedulor SLURM
 
`srun -p gpu_interactive -t 0-08:00 --mem 4G --gres=gpu:v100:1 --pty bash -i`

Check GPU available

```bash
~>lspci | grep -i nvidia
18:00.0 3D controller: NVIDIA Corporation GV100GL [Tesla V100 PCIe 32GB] (rev a1)
3b:00.0 3D controller: NVIDIA Corporation GV100GL [Tesla V100 PCIe 32GB] (rev a1)
86:00.0 3D controller: NVIDIA Corporation GV100GL [Tesla V100 PCIe 32GB] (rev a1)
af:00.0 3D controller: NVIDIA Corporation GV100GL [Tesla V100 PCIe 32GB] (rev a1)
```

**MUST** (at Annecy) with schedulor CONDOR

First **Leave your HOME** for direction in /lapp_data/... 

and ask interactive GPU session  with command

```bash
$  condor_submit -interactive interactif_job.sub
Submitting job(s).
1 job(s) submitted to cluster 19772.
Waiting for job to start...
```

Check GPU available

```
Welcome to slot1_1@lapp-wngpu002.in2p3.fr!
bash-4.2$ lspci | grep -i nvidia
84:00.0 3D controller: NVIDIA Corporation GK210GL [Tesla K80] (rev a1)
85:00.0 3D controller: NVIDIA Corporation GK210GL [Tesla K80] (rev a1)
bash-4.2$
```



### Definition of what we want to optimize

For the benchmark we will optimize the calculation time of solving the Gray-Scott equation in full definition 1920x1080 and by taking 1 image every 34 iterations. Calculations are done in float32. As the resolution on GPU is quite fast, we will keep 5000 images on GPU, ie 39 GB. We used an Nvidia Ampere 100 card with 80GB of memory for the bench.

The production of the film is outside the bench. It is created by a specific OpenCV library. In practice, we assemble images on a color/gray scale of 256 values, i.e. one byte.

So we used `datetime` module to obtain the wall time before and after Gray-Scott function and the Nvidia profiler Nsight. Concretely, the profiler increases the execution time and you have to launch it twice with and without.

```python
print(f"step_frame={step_frame}")
print(f"nb_frame={nb_frame}")
print(func_grayscott.__name__)
t_wall = datetime.now()
frames_v_ar = func_grayscott(u_ar, v_ar, Du, Dv, F, k, delta_t, nb_frame, step_frame)
duration_wall = datetime.now()-t_wall
print(f"Wall time= {duration_wall} s") 
```

### First result

```bash
1920 1080
step_frame=34
nb_frame=5000
grayscott_cupy
```

Best timing
```bash
Wall time= 0:00:43.490786 s
```

Worst timing
```bash
Wall time= 0:00:44.720730 s
```
### NVidia profiler : Nsight

`nsys profile python grayscott_cupy.py`

After open the GUI with :

`nsys-ui`

and load the report file of profilling.




### First profiling

![alt text](img/cupy_v1.png)

as expected, the calculation of the Laplacian is the main item of calculation cost. Unless we change the Laplacian calculation approach we have no room for optimization with the convol2D function.

Let's look at the memory transfer part.

![alt text](img/cupy_v1_tl2.png)

we can see the "convol+update" pattern 34 times then the copy to the CPU (DtoH). There seems to be a "dead time" (from CPU ?) between the copy to the CPU. Let's zoom in to see better.

![alt text](img/cupy_v1_tl.png)

yes there is clearly no activity for ~3 ms after copying to the CPU. Let's try to delete it by copying it to the CPU at the end of processing

### Improve memory transfert

Keep images on GPU and transfer it at the end, 3 lines to modify:

<details>
<summary>Click to expand</summary>

Replace by a DEVICE memory allocation
```python
    #frames_V = np.empty((nb_frame, n_x, n_y), dtype=np.float32)
    frames_V = cp.empty((nb_frame, n_x, n_y), dtype=cp.float32)
```

Copy DEVICE to DEVICE and return HOST array
```python
        #frames_V[idx_fr] = v_gpu.get()
        frames_V[idx_fr] = v_gpu
    #return frames_V
    return frames_V.get()
```
</details>

### Second profiling
Let's look at the Nsight timeline,

![alt text](img/cupy_v2_all.png)

 Dead time at ms level disappeared. The part of memory transfer has jumped to 50% ! Nsight was underestimated by the previous profiling, why? To many epsilon time to add , frequency sampling too low ?

Now the timing 

```bash
1920 1080
step_frame=34
nb_frame=5000
grayscott_cupy_fm2
```

Best timing
```bash
Wall time= 0:00:38.162880 s
```

Worst timing
```bash
Wall time= 0:00:40.798538 s
```
Before arround 44s, now 39s, we gain **10%** with this correction.

We were able to benefit from the elimination of dead time in the expected proportions.

### Change the dtype of array image

So the part of memory tranfert is 50%. To reduce it, we will convert array of float32 into a one-byte integer as to make the film, we should reduce the transfer time by 75% without degradation of film quality.

**We assume than V values stay between 0 and 1**, we can do that in 2 lines

```python
    frames_V *= 255
    frames_int = frames_V.astype(np.uint8)
    return frames_int.get()
```

### Thrid profiling

With transfert in uint8

![alt text](img/cupy_int8-1.png)


* The part of array convention is 2x0.3% of time kernel 

and 

* the global memory part is now of 17%.

Best timing
```bash
Wall time= 0:00:28.604349 s
```

Worst timing
```bash
Wall time= 0:00:31.436751 s
```

We saved 10 seconds

### Conclusions

* Solving the Gray-Scoot equation produces a lot of image
* The GPU vs CPU transfer time becomes significantly penalizing compared to the calculation time
* By buffering and converting images from float 32 to int8 we saved 25% on time

# Bench GPU A100@MUST for 34x1000 iterations

* For generic stencil 3x3, thanks to the convol2d() function
* 1920x1080 pixels
* 1000 images from 34x1000 iterations

```python
stl_gpu = cp.array([[1, 1, 1], 
                    [1, -8, 1], 
                    [1, 1, 1]], dtype=cp.float32)
```

```bash
step_frame=34
nb_frame=1000
grayscott_cupy_fm3
3x3 1, -8
```


```python
In [10]: from cupyx.profiler import benchmark

In [11]: print(benchmark(gsc.grayscott_main,(grayscott_cupy_fm3, gs_pars, U, V, nb_frame), n_repeat=10))

grayscott_main      :    
CPU  : 6015205.954 us   +/- 290642.948 (min: 5685796.849 / max: 6596627.789) us     
GPU-0: 6015226.367 us   +/- 290640.915 (min: 5685811.035 / max: 6596637.695) us
```
**6 seconds**