import cunumeric as np
from legate.timing import time
from gray_scott_utils import init_arrays, nrow, ncol, Du, Dv, dt, k, f, stencil
import matplotlib.pyplot as plt

def laplacian_convolve(U):
    return np.convolve(U, stencil, mode="same")

def init_arrays(nrow, ncol):
    """Create and initialize two arrays representing the concentration of the u and v species

    Args:
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created

    Returns:
        U0 (array of floats): The initialized U array
        V0 (array of floats): The initialized V array
    """

    frac = 16
    nbegin = 7
    nend = 8
    row_shift = -4
    
    U0 = np.ones((nrow, ncol), dtype=np.float32)
    V0 = np.zeros((nrow, ncol), dtype=np.float32)
    
    mask_rows_begin = int(row_shift + (nbegin*nrow)/frac)
    mask_rows_end = int(row_shift + (nend*nrow)/frac)
    mask_cols_begin = int((nbegin*ncol)/frac)
    mask_cols_end = int((nend*ncol)/frac)
    
    U0[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end] = 0
    V0[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end] = 1 
    
    return U0, V0

def gray_scott(U, V, Du, Dv, f, k, dt):
    """Propagate the u and v species in U and V arrays

    Args:
        U (array of floats): Input array of u species
        V (array of floats): Input array of v species
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate 
        k (float): Kill rate
        dt (float): Time interval between two steps

    Returns:
        U (array of floats): The updated U array
        V (array of floats): The updated V array
    """
    UVV = U * V * V
    np.add(U, (Du * laplacian_convolve(U) - UVV + f * (1 - U)) * dt, out=U)
    np.add(V, (Dv * laplacian_convolve(V) + UVV - V * (f + k)) * dt, out=V)
    
    return U, V


def simulate_images(U, V, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol):
    """Compute the Gray-Scott diffusion reaction for each time step and save images

    Args:
        U (array of floats): The initialized U array
        V (array of floats): The initialized V array
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate 
        k (float): Kill rate
        dt (float): Time interval between two steps
        nimages (int): Number of images to be created
        dsteps (int): Number of extra steps to be computed between images
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created

    Returns:
        image_array (array of floats): 3D array to store all the images
    """
    
    image_array = np.zeros((nimages, nrow, ncol))
    image_names = []    
    for idx in range(nimages):
        for step in range(dsteps) : 
            U, V = gray_scott(U, V, Du, Dv, f, k, dt)
        image_array[idx,:,:] = V 
    return image_array

nimages = 100
dsteps = 34
nrow = 1080
ncol = 1920

U0, V0 = init_arrays(nrow, ncol)
print("Start simulation")
start = time()
simulate_images(U0, V0, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol)
end = time()
print(f"Elapsed Time: {(end - start)/1000} ms")


