import os
os.environ['XLA_FLAGS'] = ( #### Write here XLA optimization flags
)

from gray_scott_utils import nrow, ncol, Du, Dv, dt, k, f

import time
import jax
import jax.numpy as jnp

def init_arrays_jax(nrow, ncol):
    """Create and initialize two JAX arrays representing the concentration of the u and v species

    Args:
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created

    Returns:
        U0_ret (array of floats): The initialized U JAX array
        V0_ret (array of floats): The initialized V JAX array
    """
    frac = 16
    nbegin = 7
    nend = 8
    row_shift = -4
    
    U0 = jnp.ones((nrow, ncol))
    V0 = jnp.zeros((nrow, ncol))
    
    mask_rows_begin = int(row_shift + (nbegin*nrow)/frac)
    mask_rows_end = int(row_shift + (nend*nrow)/frac)
    mask_cols_begin = int((nbegin*ncol)/frac)
    mask_cols_end = int((nend*ncol)/frac)
    
    U0_ret = U0.at[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end].set(0) 
    V0_ret = V0.at[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end].set(1) 
    
    return U0_ret, V0_ret

@jax.jit
def laplacian_unroll_jax(u):
    """Applies a 3x3 stencil and fills the border with 0
       1, 1,1
       1,-8,1
       1, 1,1
       
    Args:
        U (array of floats): The array on which we will apply the stencil

    Returns:
        lap_full: The laplacian of U
    """
    
    lap = u.at[:-2, 1:-1].get()
    lap = lap + u.at[1:-1, :-2].get()
    lap = lap + u.at[1:-1, 2:].get()
    lap = lap + u.at[2:, 1:-1].get()
    lap = lap - 8.0 * u.at[1:-1, 1:-1].get()
    lap = lap + u.at[:-2, :-2].get()
    lap = lap + u.at[2:, 2:].get()
    lap = lap + u.at[2:, :-2].get()
    lap = lap + u.at[:-2, 2:].get()
    lap_full = jnp.zeros_like(u)
    return lap_full.at[1:-1, 1:-1].set(lap)

@jax.jit
def gray_scott_jax(U, V, Du, Dv, f, k, dt):
    """Propagate the u and v species in U and V arrays

    Args:
        U (array of floats): Input array of u species
        V (array of floats): Input array of v species
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate 
        k (float): Kill rate
        dt (float): Time interval between two steps

    Returns:
        U (array of floats): The updated U array
        V (array of floats): The updated V array
    """
    UVV = jnp.zeros((len(U), len(U[0])))

    UVV = UVV.at[:,:].set(U * V * V)
    U = U.at[:,:].add((Du * laplacian_unroll_jax(U) - UVV + f * (1 - U)) * dt)
    V = V.at[:,:].add((Dv * laplacian_unroll_jax(V) + UVV - V * (f + k)) * dt)
    
    return U, V

def simulate_images_jax(U, V, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol):
    """Compute the Gray-Scott diffusion reaction for each time step and save images

    Args:
        U (array of floats): The initialized U array
        V (array of floats): The initialized V array
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate 
        k (float): Kill rate
        dt (float): Time interval between two steps
        nimages (int): Number of images to be created
        dsteps (int): Number of extra steps to be computed between images
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created


    Returns:
        image_array (array of floats): 3D array to store all the images
    """
    image_array = jnp.zeros((nimages, nrow, ncol))
        
    for idx in range(nimages):
        for step in range(dsteps) : 
            U, V = gray_scott_jax(U, V, Du, Dv, f, k, dt)
        image_array = image_array.at[idx,:,:].set(V)
            
    return image_array 

def main():
    #### Use these functions to compute on GPU for 34 x 1000 images
    return 

if __name__=='__main__':
    main()
