import os
os.environ['XLA_FLAGS'] = (
    '--xla_gpu_enable_triton_softmax_fusion=true '
    '--xla_gpu_triton_gemm_any=True '
)

from gray_scott_utils import nrow, ncol, Du, Dv, dt, k, f

import time
import jax
import jax.numpy as jnp

def init_arrays_jax(nrow, ncol):
    """Create and initialize two JAX arrays representing the concentration of the u and v species

    Args:
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created

    Returns:
        U0_ret (array of floats): The initialized U JAX array
        V0_ret (array of floats): The initialized V JAX array
    """
    frac = 16
    nbegin = 7
    nend = 8
    row_shift = -4
    
    U0 = jnp.ones((nrow, ncol))
    V0 = jnp.zeros((nrow, ncol))
    
    mask_rows_begin = int(row_shift + (nbegin*nrow)/frac)
    mask_rows_end = int(row_shift + (nend*nrow)/frac)
    mask_cols_begin = int((nbegin*ncol)/frac)
    mask_cols_end = int((nend*ncol)/frac)
    
    U0_ret = U0.at[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end].set(0) 
    V0_ret = V0.at[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end].set(1) 
    
    return U0_ret, V0_ret

@jax.jit
def laplacian_unroll_jax(u):
    """Applies a 3x3 stencil and fills the border with 0
       1, 1,1
       1,-8,1
       1, 1,1
       
    Args:
        U (array of floats): The array on which we will apply the stencil

    Returns:
        lap_full: The laplacian of U
    """
    
    lap = u.at[:-2, 1:-1].get()
    lap = lap + u.at[1:-1, :-2].get()
    lap = lap + u.at[1:-1, 2:].get()
    lap = lap + u.at[2:, 1:-1].get()
    lap = lap - 8.0 * u.at[1:-1, 1:-1].get()
    lap = lap + u.at[:-2, :-2].get()
    lap = lap + u.at[2:, 2:].get()
    lap = lap + u.at[2:, :-2].get()
    lap = lap + u.at[:-2, 2:].get()
    lap_full = jnp.zeros_like(u)
    return lap_full.at[1:-1, 1:-1].set(lap)

@jax.jit
def gray_scott_jax(U, V, Du, Dv, f, k, dt):
    """Propagate the u and v species in U and V arrays

    Args:
        U (array of floats): Input array of u species
        V (array of floats): Input array of v species
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate 
        k (float): Kill rate
        dt (float): Time interval between two steps

    Returns:
        U (array of floats): The updated U array
        V (array of floats): The updated V array
    """
    UVV = jnp.zeros((len(U), len(U[0])))

    UVV = U * V * V
    U = U + (Du * laplacian_unroll_jax(U) - UVV + f * (1 - U)) * dt
    V = V + (Dv * laplacian_unroll_jax(V) + UVV - V * (f + k)) * dt
    
    return U, V


@jax.jit
def gray_scott_jax(U, V, Du, Dv, f, k, dt):
    UVV = U * V * V
    U = U + (Du * laplacian_unroll_jax(U) - UVV + f * (1 - U)) * dt
    V = V + (Dv * laplacian_unroll_jax(V) + UVV - V * (f + k)) * dt
    return U, V
    

@partial(jax.jit, static_argnames=['nimages', 'dsteps'])
def simulate_images_jax(U, V, Du, Dv, f, k, dt, *, nimages, dsteps, nrow, ncol):
    def simulate_dsteps(carry, _):
        U, V = carry

        def step_fn(i, uv):
            U, V = uv
            return gray_scott_jax(U, V, Du, Dv, f, k, dt)

        U, V = jax.lax.fori_loop(0, dsteps, step_fn, (U, V))
        return (U, V), V

    (U_final, V_final), image_array = jax.lax.scan(simulate_dsteps, (U, V), None, length=nimages)
    return image_array

def main():
    start = time.time()
    nimages = 1000
    U0, V0 = init_arrays_jax(nrow, ncol)
    simulate_images_jax(U0, V0, Du, Dv, f, k, dt, nimages).block_until_ready()
    end = time.time()
    print(f"Simulation time : {end-start} s") 
    return 

if __name__=='__main__':
    main()


