import numpy as np
import matplotlib.pyplot as plt
import cv2
import time

# Custom stencil 
stencil = np.array([[1., 1., 1.],[1., -8., 1.],[1., 1., 1.]])

# Size of the images
nrow = 1080
ncol = 1920

# Number of images (keep it low for the moment and we will increase it when we have a faster program)
nimages = 10

# Number of extra steps (same here: keep it low !)
dsteps = 2

# Update in time
dt = 1.0

# Diffusion coefficients
Du = 0.1
Dv = 0.05

# Define feed/kill rates
k = 0.054
f = 0.014


def init_arrays(nrow, ncol):
    """Create and initialize two arrays representing the concentration of the u and v species

    Args:
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created

    Returns:
        U0 (array of floats): The initialized U array
        V0 (array of floats): The initialized V array
    """
    frac = 16
    nbegin = 7
    nend = 8
    row_shift = -4
    
    U0 = np.ones((nrow, ncol), dtype=np.float32)
    V0 = np.zeros((nrow, ncol), dtype=np.float32)
    
    mask_rows_begin = int(row_shift + (nbegin*nrow)/frac)
    mask_rows_end = int(row_shift + (nend*nrow)/frac)
    mask_cols_begin = int((nbegin*ncol)/frac)
    mask_cols_end = int((nend*ncol)/frac)
    
    U0[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end] = 0
    V0[mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end] = 1 
    
    return U0, V0


def simulate_images(U, V, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol, stencil):
    """Compute the Gray-Scott diffusion reaction for each time step and save images

    Args:
        U (array of floats): The initialized U array
        V (array of floats): The initialized V array
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate 
        k (float): Kill rate
        dt (float): Time interval between two steps
        nimages (int): Number of images to be created
        dsteps (int): Number of extra steps to be computed between images
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created
        stencil (array of floats): the stencil describing the propagation

    Returns:
        image_array (array of floats): 3D array to store all the images
    """
    
    image_array = np.zeros((nimages, nrow, ncol))
        
    for idx in range(nimages):
        for step in range(dsteps) : 
            U, V = gray_scott(U, V, Du, Dv, f, k, dt, stencil)
        image_array[idx,:,:] = V
            
    return image_array  


def save_images(image_array):
    """Save the images created for each time step 

    Args:
        image_array (array of floats): 3D array with all the images to save

    Returns:
        image_names (list of str): List of the names of the images 
    """
    
    image_names=[]
    
    for idx in range(len(image_array)):
        name = 'results/'+'img'+str(idx)+'.png' 
        plt.imsave(name, image_array[idx])
        image_names.append(name)
        
    return image_names


def make_video(image_names):
    """Create a video from a list of image names

    Args:
        image_names (array of str): The list of the image names
    """
    
    img_array = []
    
    for name in image_names: 
        img = cv2.imread(name)
        height, width, layers = img.shape
        size = (width,height)
        img_array.append(img)
 
    out = cv2.VideoWriter('results/project.avi',cv2.VideoWriter_fourcc(*'DIVX'), fps=10, frameSize=size)

    for i in range(len(img_array)):
        out.write(img_array[i])
    out.release()
    
    return


