# -*- coding: utf-8 -*-
"""1_Python_PytorchConv2d_GrayScott.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/1XhAw4dpZPGTS1dyGF4eTfZKh9FIsJXZR

# Performance in Python

Python is used most of the time as an interpreted language. If so, it is not compiled to machine code like C, Fortran, etc. Instead, the source code is compiled to *byte code*, an intermediate format which is then read and executed line by line by the interpreter. So, Python code is slower than fully compiled code, but more flexible and easy to use.
In addition, Python is dynamically typed : the user does not have to specify the type of the variable they declare.

Most computing libraries used in Python are written in C and compiled (`numpy`, `math`, etc), and the only slowdown in execution is when they are called through the Python code.
An easy way to have a fast Python code is then to spend most of the computing time using these optimized libraries.

It is also possible to use compilers for subsets of the Python language such as *Pythran* and *Numba* to speed up the Python code.

In this class we will learn how to write fast Python code, and further optimize it for CPU. We will take the example of the Gray-Scott diffusion reaction.

## Implementation of the Gray-Scott reaction

### General equations

The diffusion reaction involves three species **U**, **V** and **P** such that :

\begin{split}
\begin{array}{l}
\displaystyle U + 2V \rightarrow 3V \\
\displaystyle V \rightarrow P \\
\end{array}
\end{split}

**U** is added at a given *feed rate* $r_u$, **V** uses **U** to grow, and is removed (transformed to **P**) at a given *kill rate* $r_v$ :

\begin{split}
\begin{array}{l}
\displaystyle \frac{\partial u}{\partial t} = r_u \Delta u - uv^2 + f_r(1-u) \\
\displaystyle \frac{\partial v}{\partial t} = r_v \Delta v + uv^2 - (f_r+k_r)v
\end{array}
\end{split}

where $u$ and $v$ are the concentration of **U** and **V** species.


The reaction naturally has two stable states: lots of **U** with no **V** to consume it, or lots of **V** where new **U** is quickly converted into more **V**. But the diffusion can cause interesting behaviors at the borders between the mostly-**U** and mostly-**V** areas.

$\Delta u$ is the laplacian of u. It measures the difference between $u$ at a point and the average of $u$ around this point.

We will create 2D grids where we will compute $u$ and $v$ at each cell. The result will be a succession of images describing the evolution of the concentrations.

### Stencil implementation

On our grid, we can discretize the Laplacian, and compute it for each cell $(i, j)$ by summing the differences between the neighbouring cells and the cell $(i,j)$.

#### 1. Von Neumann neighbourhood
If we consider that **U** only diffuses to its von Neumann neighbourhood (i.e. only the four closest neighbours and not the diagonal neighbours), the discretized Laplacian is of this form :

\begin{split}
\begin{array}{l}
\displaystyle \Delta u_{i,j} = u_{i,j-1} - u_{i,j} + u_{i-1,j} - u_{i,j} + u_{i+1, j} - u_{i,j} + u_{i, j+1} - u_{i,j} \\
\displaystyle \Delta u_{i,j} = u_{i,j-1} + u_{i-1,j} + u_{i+1, j} + u_{i, j+1} - 4u_{i,j}
\end{array}
\end{split}

To implement this computation, we can define a convolution stencil :

\begin{align}
h=\begin{bmatrix} 0 & 1 & 0 \\ 1 & -4 & 1 \\ 0 & 1 & 0\end{bmatrix}
\end{align}

Indeed, if we compute the convolution product of this stencil with $u$ :

\begin{split}
\begin{array}{l}
\displaystyle(u\ast h)(i, j) = \sum_{k,l=0}^{k,l = N-1} u_{i-k, j-l}h_{k,l} \\
\displaystyle(u\ast h)(i, j) = u_{i,j-1} + u_{i-1,j} + u_{i+1, j} + u_{i, j+1} - 4u_{i,j} \\
\displaystyle(u\ast h)(i, j) = \Delta u_{i,j}
\end{array}
\end{split}

with $N$ being the size of our square stencil (here $N=3$), we find the Laplacian of *u* at cell $(i,j)$.

#### 2. Moore neighbourhood
Another possibility of disctretization is with a stencil representing the diffusion of **U** to its Moore neighbourhood (i.e. the eight neighbours, including the diagonal ones). In this case, the stencil is of this form :

\begin{align}
h=\begin{bmatrix} 1 & 1 & 1 \\ 1 & -8 & 1 \\ 1 & 1 & 1\end{bmatrix}
\end{align}

To compare with the other implementations in C++, Rust and Sycl, we choose to use this 9-elements stencil for the Python implementation as well.

## Pytorch Version of GrayScott - JMC
"""

import numpy as np
#import matplotlib.pyplot as plt
#import cv2
import os
#from scipy import signal
import timeit

"""Import the necessay pytorch libraries
Pytorch functional is a pytorch sub-library dedicated to tensor operations
"""

import torch
import torch.nn.functional as F

torch.cuda.current_device()

torch.cuda.get_device_name(0)

"""Stencil (laplacian convolution declaration) : must be a tensor for 2D convolution : make attention to dimensions"""

stencil = np.zeros((1,1,3,3), dtype=np.float32)
stencil[0,0,:,:] = np.array([[1., 1., 1.],[1., -8., 1.],[1., 1., 1.]])
stencil = torch.tensor(stencil)
print(stencil)

"""We will convert all the precedent variable to Torch tensors along with the functions"""

def gray_scott_conv2d_pt(U, V, Du, Dv, f, k, dt, stencil):
    """Propagate the u and v species in U and V arrays

    Args:
        U (array of floats): Input array of u species
        V (array of floats): Input array of v species
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate
        k (float): Kill rate
        dt (float): Time interval between two steps
        stencil (array of floats): the stencil describing the propagation

    Returns:
        U (array of floats): The updated U array
        V (array of floats): The updated V array
    """
    # Load variables to GPU
    U = U.cuda()
    V = V.cuda()
    stencil = stencil.cuda()
    f= f.cuda()
    k= k.cuda()
    dt= dt.cuda()

    # calculation

    UVV = U * V * V

    lapU = F.conv2d(U, stencil, stride=1, padding=1)
    lapV = F.conv2d(V, stencil, stride=1, padding=1)

    U += (Du * lapU - UVV + f * (1 - U)) * dt
    V += (Dv * lapV + UVV - V * (f + k)) * dt

    # Move variables back to CPU
    #U = U.cpu()
    #V = V.cpu()

    #U=U.numpy()
    #V=V.numpy()

    #U= U[0,0]
    #V= V[0,0]

    return U, V

def simulate_images_conv2d_pt(U, V, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol, stencil):
    """Compute the Gray-Scott diffusion reaction for each time step and save images

    Args:
        U (array of floats): The initialized U array
        V (array of floats): The initialized V array
        Du (float): Diffusion rate of the u species
        Dv (float): Diffusion rate of the v species
        f (float): Feed rate
        k (float): Kill rate
        dt (float): Time interval between two steps
        nimages (int): Number of images to be created
        dsteps (int): Number of extra steps to be computed between images
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created
        stencil (array of floats): the stencil describing the propagation

    Returns:
        image_array (array of floats): 3D array to store all the images
    """

    image_array = np.zeros((nimages, nrow, ncol))

    for idx in range(nimages):
        for step in range(dsteps) :
            U, V = gray_scott_conv2d_pt(U, V, Du, Dv, f, k, dt, stencil)
        # ATTETION : Move V tensor to CPU before converting to NumPy array
        image_array[idx,:,:] = V.cpu().numpy()

    return image_array

"""Make sure to separate the variables dedicated to computing on GPU, from variable dedicated to computing on CPU (essentially for the loop handelin)"""

def init_arrays_pt(nrow, ncol):
    """Create and initialize two arrays representing the concentration of the u and v species

    Args:
        nrow (int): Number of rows of the arrays to be created
        ncol (int): Number of columns of the arrays to be created

    Returns:
        U0 (array of floats): The initialized U array
        V0 (array of floats): The initialized V array
    """
    frac = 16
    nbegin = 7
    nend = 8
    row_shift = -4

    #U0 = torch.ones((1,1,nrow, ncol), dtype=torch.float32, device=torch.device('cpu'))
    #V0 = torch.zeros((1,1,nrow, ncol), dtype=torch.float32, device=torch.device('cpu'))

    # in case the GPU runtime is not activated : the following expression avoids the error of variable declaration on the GPU

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    U0 = torch.ones((1,1,nrow, ncol), dtype=torch.float32, device=device)
    V0 = torch.zeros((1,1,nrow, ncol), dtype=torch.float32, device=device)

    # Mask creation to populate predators

    mask_rows_begin = int(row_shift + (nbegin*nrow)/frac)
    mask_rows_end = int(row_shift + (nend*nrow)/frac)
    mask_cols_begin = int((nbegin*ncol)/frac)
    mask_cols_end = int((nend*ncol)/frac)

    U0[0,0,mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end] = 0
    V0[0,0,mask_rows_begin:mask_rows_end, mask_cols_begin:mask_cols_end] = 1

    return U0, V0

def save_images_pt(image_array):
    """Save the images created for each time step

    Args:
        image_array (array of floats): 3D array with all the images to save

    Returns:
        image_names (list of str): List of the names of the images
    """

    image_names=[]

        # Create the 'results' directory if it doesn't exist
    if not os.path.exists('results'):
        os.makedirs('results')

    for idx in range(len(image_array)):
        name = 'results/'+'img'+str(idx)+'.png'
        plt.imsave(name, image_array[idx])
        image_names.append(name)

    return image_names

def make_video_pt(image_names):
    """Create a video from a list of image names

    Args:
        image_names (array of str): The list of the image names
    """

    img_array = []

    for name in image_names:
        img = cv2.imread(name)
        height, width, layers = img.shape
        size = (width,height)
        img_array.append(img)

    out = cv2.VideoWriter('results/project.avi',cv2.VideoWriter_fourcc(*'DIVX'), fps=10, frameSize=size)

    for i in range(len(img_array)):
        out.write(img_array[i])
    out.release()

    return

"""Declare model parameter : make attention to variables dedicated to loop managment and variables dedicated to GPU computing"""

# Size of the images
nrow = 540
ncol = 960

# Number of images (keep it low for the moment and we will increase it when we have a faster program)
nimages = 100

# Number of extra steps (same here: keep it low !)
dsteps = 2

# Update in time
dt = torch.tensor([1.0], device=torch.device('cuda' if torch.cuda.is_available() else 'cpu'))

# Diffusion coefficients
Du = torch.tensor([0.1], device=torch.device('cuda' if torch.cuda.is_available() else 'cpu'))
Dv = torch.tensor([0.05], device=torch.device('cuda' if torch.cuda.is_available() else 'cpu'))

# Define feed/kill rates
k = torch.tensor([0.054], device=torch.device('cuda' if torch.cuda.is_available() else 'cpu'))
f = torch.tensor([0.014], device=torch.device('cuda' if torch.cuda.is_available() else 'cpu'))

U0, V0 = init_arrays_pt(nrow, ncol)
#plt.imshow(V0[0, 0].cpu()) # Show initial concentrations

import datetime
t_wall = datetime.now()

image_array = simulate_images_conv2d_pt(U0, V0, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol, stencil)

duration_wall = datetime.now()-t_wall
print(f"Wall time= {duration_wall} s")



#image_names = save_images_pt(image_array)
#make_video_pt(image_names)

# Commented out IPython magic to ensure Python compatibility.
# %timeit simulate_images_conv2d_pt(U0, V0, Du, Dv, f, k, dt, nimages, dsteps, nrow, ncol, stencil)

"""Not thar fast as expected , now let's have fan comparing and optimizing : change image size and compare with scikit convolution

# Take away problems for next Gary-Scott Revolution Session:

Modify the Image array to be handeled on GPU with image sequence added on the GPU


Make the complete video on the GPU  to avoid image transfer to cpu , by using torchvision library

Use Pytorch JIT library for compilation
"""

